/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
   There are four ways an object can appear larger in an image.
   The object can grow in size (scale up), the object can move
   closer to the camera, the camera can move closer to the object,
   or the camera can zoom in on the object.
<p>
   If we look at an object without any frame of reference, we
   cannot tell which of these four operations is making the
   object larger (this ambiguity is the basis for many optical
   illusions). But the right frame of reference can give us
   clues about which operation is causing the object to appear
   larger.
*/
public class Zoom_vs_Scale_vs_Move
{
   private static boolean showMatrix = false;
   private static double zTranslation = 0.0;
   private static double eyeZ = 2.0;
   private static double eyeY = 0.0;
   private static double scale = 1.0;
   private static double zoom = 1.0;

   private static double near   =  1.0;

   private static void print_help_message()
   {
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the m/M, key to move the sphere along the z axe.");
      System.out.println("Use the up/down arrow keys to move the camera along the z axe.");
      System.out.println("Use the s/S keys to scale the size of the sphere.");
      System.out.println("Use the z/Z keys to zoom the camera in and out of the scene.");
      System.out.println("Use the 'f' key to toggle the display of the floor & wall on and off.");
      System.out.println("Use the 'c' key to change the random sphere color.");
      System.out.println("Use the 'v' key to toggle showing the Model, View, & Normalize matrices.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   public static void main(String[] args)
   {
      print_help_message();

      Model sphere = new Sphere(1.0, 30, 30);
      sphere.setColor(Color.red);
      Position sphere_p = new Position(sphere);

      Model xzPlane = new Panel(-20, 20, -15, 0);
      xzPlane.setColor(Color.lightGray);
      Position xzPlane_p = new Position(xzPlane);
      xzPlane_p.translate(0, -1, 0);

      Model xyPlane = new Panel(-20, 20, -15, 1);
      xyPlane.setColor(Color.lightGray);
      Position xyPlane_p = new Position(xyPlane);
      xyPlane_p.rotateX(90);
      xyPlane_p.translate(0, -15, 0);

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the models to the Scene (the order is important).
      scene.addPosition(xzPlane_p, xyPlane_p, sphere_p);

      // Set up the camera's initial view volume.
      scene.camera.projPerspective(-1, +1, -1, +1, near);

      // Set up the camera's initial location.
      scene.camera.viewTranslate(0, 0, eyeZ);

      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      InteractiveFrame app = new InteractiveFrame("Zoom_vs_Scale_vs_Move", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyPressed(KeyEvent e)
         {
            //System.out.println( e );

            int keyCode = e.getKeyCode();
            // Only handle the four arrow keys.
            if (KeyEvent.VK_UP == keyCode   || KeyEvent.VK_DOWN == keyCode
             || KeyEvent.VK_LEFT == keyCode || KeyEvent.VK_RIGHT == keyCode)
            {
               if (KeyEvent.VK_UP == keyCode && e.getModifiers() != 0 )
               {
                  eyeY += 0.1;
               }
               else if (KeyEvent.VK_DOWN == keyCode && e.getModifiers() != 0 )
               {
                  eyeY -= 0.1;
               }
               if (KeyEvent.VK_UP == keyCode)
               {
                  eyeZ += 0.1;
               }
               else if (KeyEvent.VK_DOWN == keyCode)
               {
                  eyeZ -= 0.1;
               }

               // Set up the camera's location and orientation.
               scene.camera.view2Identity();
               scene.camera.viewTranslate(0, eyeY, eyeZ);

               if (showMatrix)
               {
                  System.out.println("The Sphere's Model matrix");
                  System.out.print( sphere_p.modelMatrix );
                  System.out.println("The Camera's View matrix");
                  System.out.print( scene.camera.viewMatrix );
                  System.out.println("The Camera's Normalize matrix");
                  System.out.println( scene.camera.normalizeMatrix );
               }

               // Render again.
               FrameBuffer fb = this.fbp.getFrameBuffer();
               fb.clearFB(Color.black);
               Pipeline.render(scene, fb);
               fbp.update();
               repaint();
            }
         }


         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );

            char c = e.getKeyChar();
            if ('f' == c)
            {
               xzPlane.visible = ! xzPlane.visible;
               xyPlane.visible = ! xyPlane.visible;
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('v' == c)
            {
               showMatrix = ! showMatrix;
            }
            else if ('m' == c)
            {
               zTranslation += 0.1;
            }
            else if ('M' == c)
            {
               zTranslation -= 0.1;
            }
            else if ('s' == c) // Scale the sphere 5% smaller.
            {
               scale /= 1.05;
            }
            else if ('S' == c) // Scale the sphere 5% larger.
            {
               scale *= 1.05;
            }
            else if ('z' == c)
            {
               near -= 0.01;
            }
            else if ('Z' == c)
            {
               near += 0.01;
            }
            else if ('c' == c)
            {
               // Change the color of the sphere.
               sphere.setColorRandom();
            }

            // Set the model-to-view transformation matrix.
            // The order of the transformations is important!
            sphere_p.matrix2Identity();
            sphere_p.translate(0, 0, zTranslation);
            sphere_p.scale(scale);



      // Set up the camera's view volume.
      if (scene.camera.perspective)
      {
         scene.camera.projPerspective(-1, +1, -1, +1, near);
      }
      else
      {
         scene.camera.projOrtho(-1, +1, -1, +1);
      }

            if (showMatrix && ('v'==c||'s'==c||'m'==c||'z'==c||
                               'p'==c||'S'==c||'M'==c||'Z'==c))
            {
               System.out.println("The Sphere's Model matrix");
               System.out.print( sphere_p.modelMatrix );
               System.out.println("The Camera's View matrix");
               System.out.print( scene.camera.viewMatrix );
               System.out.println("The Camera's Normalize matrix");
               System.out.println( scene.camera.normalizeMatrix );
            }

            // Render again.
            FrameBuffer fb = this.fbp.getFrameBuffer();
            fb.clearFB(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );

            // Get the new size of the FrameBufferPanel.
            int w = this.fbp.getWidth();
            int h = this.fbp.getHeight();

            // Create a new FrameBuffer that fits the new window size.
            FrameBuffer fb = new FrameBuffer(w, h);
            this.fbp.setFrameBuffer(fb);
            fb.clearFB(Color.darkGray);

            // Create a maximal viewport with aspect ratio 1
            // in the center of the FrameBuffer.
            if ( 1.0 <= w/(double)h )
            {
               int xOffset = (w - h)/2;
               fb.setViewport(xOffset, 0, h, h);
            }
            else
            {
               int yOffset = (h - w)/2;
               fb.setViewport(0, yOffset, w, w);
            }
            fb.clearFB(Color.darkGray);
            fb.clearVP(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }
      };
   }
}
