/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer} data structure. This renderer mutates
   the {@link FrameBuffer}'s current viewport so that it is filled
   in with the rendered image of the scene represented by the
   {@link Scene} object.
<p>
   This implements our tenth rendering pipeline. In the previous
   renderer the first pipeline stage was Model2View. This renderer
   splits that stage into two stages, {@link Model2World} and
   {@link World2View}. The reason for the new stage is that we have
   a new coordinate system, world coordinates, that is logically
   between model coordinates and view coordinates.  Recall that
   "model coordinates" are relative to a model's private origin
   and "view coordinates" are relative to the camera's arbitrary
   view volume. World coordinates are relative to a fixed origin
   shared by everything in the scene, including the camera. In this
   renderer the camera can be placed anywhere in the world coordinate
   system. In other words, the world coordinate system facilitates
   moving the {@link Camera} around in a scene. There are now seven
   pipeline stages.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position(scene, position2, Matrix.identity(), fb);
         }
      }
   }


   /**
      Recursively renderer a {@link Position} without any debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 1. Apply the current model-to-world coordinate transformation.
         Model2World.model2world(position.model.vertexList, ctm);

         // 2. Apply the camera's world-to-view coordinate transformation.
         World2View.world2view(position.model.vertexList, scene.camera.viewMatrix);

         // 3. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(position.model.vertexList, scene.camera);

         // 4. Apply the projection transformation.
         Projection.project(position.model.vertexList, scene.camera);

         // 5. Clip each line segment to the standard view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : position.model.lineSegmentList)
         {
            if ( Clip.clip(ls) )
            {
               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
         }
         position.model.lineSegmentList = lineSegmentList2;

         // 6. Apply viewport transformation to each visible line segment.
         // (Alternatively, we could apply viewport transformation to every vertex?)
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (LineSegment ls : position.model.lineSegmentList)
         {
            Viewport.viewport(ls, transformed, fb);
         }

         // 7. Rasterize each visible line segment into pixels.
         for (LineSegment ls : position.model.lineSegmentList)
         {
            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls, fb);
            }
            else
            {
               Rasterize.rasterize(ls, fb);
            }
         }

         // Recursively render every nested Position of this Model.
         for (Position p : position.model.nestedPositions)
         {
            if ( p.visible )
            {
               // Do a (pre-order) depth-first-traversal from this nested Position.
               render_position(scene, p, ctm, fb);
            }
         }
      }

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position(scene, p, ctm, fb);
         }
      }
   }



   /**
      Renderer the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         System.out.println("=== Render Position ================================");
         if ( position.visible )
         {
            // Make a deep copy of this Position.
            Position position2 = new Position(position);

            // Begin a pre-order depth-first-traversal from this Position's copy.
            render_position_debug(scene, position2, Matrix.identity(), fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("=== End Position ================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a {@link Position} with quite a bit of debugging information.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position_debug(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      System.out.println("=== Render Model ===================================");

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // log interesting information to standard output
         logVertexList("0. Model    ", position.model.vertexList);

         // 1. Apply the current model-to-world coordinate transformation.
         Model2World.model2world(position.model.vertexList, ctm);

         // log interesting information to standard output
         logVertexList("1. World", position.model.vertexList);

         // 2. Apply the camera's world-to-view coordinate transformation.
         World2View.world2view(position.model.vertexList, scene.camera.viewMatrix);

         // log interesting information to standard output
         logVertexList("2. View", position.model.vertexList);

         // 3. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(position.model.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("3. Camera", position.model.vertexList);

         // 4. Apply the projection transformation.
         Projection.project(position.model.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("4. Projected", position.model.vertexList);

         // 5. Clip each line segment to the standard view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : position.model.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("5. Clipping", ls);

            if ( Clip.clip(ls) )
            {
               // log interesting information to standard output
               logLineSegment("5. Clipped (accept)", ls);

               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
            else
            {
               // log interesting information to standard output
               logLineSegment("5. Clipped (reject)", ls);
            }
         }
         position.model.lineSegmentList = lineSegmentList2;

         // log interesting information to standard output
         logVertexList("5. Clipped  ", position.model.vertexList);

         // 6. Apply viewport transformation to each visible line segment.
         boolean[] transformed = new boolean[position.model.vertexList.size()];
         for (LineSegment ls : position.model.lineSegmentList)
         {
            Viewport.viewport(ls, transformed, fb);

            // log interesting information to standard output
            logLineSegment("6. Viewport ", ls);
         }

         // 7. Rasterize each visible line segment into pixels.
         for (LineSegment ls : position.model.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("7. Rasterize", ls);

            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls, fb);
            }
            else
            {
               Rasterize.rasterize(ls, fb);
            }
         }
         System.out.println();

         // Recursively render every nested Position of this Model.
         for (Position p : position.model.nestedPositions)
         {
            System.out.println("====== Render Model's Nested Position ================");
            if ( p.visible )
            {
               // Do a pre-order depth-first-traversal from this nested Position.
               render_position(scene, p, ctm, fb);
            }
            else
            {
               System.out.println("   Hidden position.");
            }
            System.out.println("====== End Model's Nested Position ===================");
         }
      }
      else
      {
         System.out.println("   Hidden model.");
      }

      System.out.println("=== End Model ===================================");

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         System.out.println("====== Render Nested Position ========================");
         if ( p.visible )
         {
            // Do a pre-order depth-first-traversal from this nested Position.
            render_position_debug(scene, p, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("====== End Nested Position ===========================");
      }
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
      int index0 = ls.index[0];
      int index1 = ls.index[1];
      Vertex v0 = ls.model.vertexList.get(index0);
      Vertex v1 = ls.model.vertexList.get(index1);
      System.out.printf("    index = %3d, " + v0.toString(), index0);
      System.out.printf("                 " + v0.toStringColor());
      System.out.printf("    index = %3d, " + v1.toString(), index1);
      System.out.printf("                 " + v1.toStringColor(), index1);
   }
}
